/*
 *  INTERPRETER, {akii-root}/src/interpreter.c
 *
 * ==========================================================================
 *
 *                      Arena Kliens II. Releases
 *
 * ==========================================================================
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A legkiralyabb szovegszerkeszto: VIM - Vi IMproved 5  <http://www.vim.org>
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include "interpreter.h"
#include "error.h"
#include "variables.h"
#include "expression.h"
#include "commands.h"
#include "regexp.h"
#include "screen.h"
#include "main.h"
#include "debug.h"

int sor = 0, oszlop = 0;
int msor = 0, moszlop = 0;
int inloop = 0, inkapocs = 0, inalias = 0, ident = 0;
int incheck = 0, inon = 0;
int brkcnt_level = 0, checkpoint_level = 0, nexton_level = 0;
char *mp;

struct valtozo *akii_novice = NULL;
int novice = 1;

void mso(char *p)
{
    msor = sor;
    moszlop = oszlop;
    mp = p;
    return;
}
void resetsoroszlop(int isor)
{
    sor = isor;
    oszlop = 1;
    inkapocs = 0;
    inloop = 0;
    inalias = 0;
    incheck = 0;
    inon = 0;
    mso(NULL);
    return;
}

int inc(char **p)
{
    if (**p == 10) {
	if (sor) {
	    sor++;
	    oszlop = 1;
	}
    }
    else
	oszlop++;
    (*p)++;
}
void seterrtemp(void)
{
    if (!msor)
	sprintf(errtemp, "Hiba a(z) %d. karakternel", moszlop);
    else
	sprintf(errtemp, "hiba a(z) %d. sor, %d. oszlopanal", msor, moszlop);
    return;
}
void *freepar(struct parancs *par)
{
    struct onstruct *os;
    struct wstruct *iew;
    struct ifstruct *ifs;
    struct foreachstruct *fes;
    struct forstruct *fs;
    struct timerstruct *ts;
    struct aliasstruct *ali;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_freepar, "?");
#endif
    if (!par) {
#ifdef AKII_DEBUG
	dbgleave(dbg_freepar);
#endif
	return (NULL);
    }
    par->refnum--;
    if (par->refnum > 0) {
#ifdef AKII_DEBUG
	dbgleave(dbg_freepar);
#endif
	return (NULL);
    }
    if (par->next)
	freepar(par->next);
    if (par->ertek)
	switch (par->tipus) {
	    case PAR_KIF:
		free(par->ertek);
		break;
	    case PAR_ONSEND:
	    case PAR_ON:
		os = par->ertek;
		if (os->nev)
		    free(os->nev);
		for (; os->kifek; os->kifek--) {
		    free(os->regkifek[os->kifek - 1]);
		    if (os->kifek == 1)
			free(os->regkifek);
		}
		freepar(os->par);
		free(os);
		break;
	    case PAR_IF:
		ifs = par->ertek;
		freekiffa(ifs->kif);
		freepar(ifs->par);
		freepar(ifs->elsepar);
		free(ifs);
		break;
	    case PAR_DO:
	    case PAR_WHILE:
		iew = par->ertek;
		freekiffa(iew->kif);
		freepar(iew->par);
		free(iew);
		break;
	    case PAR_FOR:
		fs = par->ertek;
		for (i = 0; i < 3; i++)
		    freekiffa(fs->kifek[i]);
		freepar(fs->par);
		free(fs);
		break;
	    case PAR_FOREACH:
		fes = par->ertek;
		free(fes->tomb);
		free(fes->inev);
		break;
	    case PAR_TIMER:
		ts = par->ertek;
		if (ts->nev)
		    free(ts->nev);
		if (ts->tipus == '1' || ts->tipus == '2') {
		    free(ts->ido);
		}
		break;
	    case PAR_RETURN:
		if (par->ertek)
		    freekiffa(par->ertek);
		break;
	    case PAR_ALIAS:
		ali = par->ertek;
		free(ali->nev);
		freepar(ali->par);
		for (i = 0; i < ali->cpnum; i++)
		    free(ali->cps[i]);
		if (ali->cps)
		    free(ali->cps);
		free(ali);
		break;
	}
    free(par);
#ifdef AKII_DEBUG
    dbgleave(dbg_freepar);
#endif
    return (NULL);
}
struct parancs *read_timer(char **s)
{
    struct parancs *par;
    struct timerstruct *ts;
    char *p = *s, *q = *s;
    char temp[256], *t = temp;
    REXELEM szavak =
    {0, NULL};
    int i, j;
    char *nev = NULL;
    int tipus = '~';
    char *idostr = NULL;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_timer, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_timer);
#endif
	return (NULL);
    }
    strncpy(temp, q, p - q);
    temp[p - q] = 0;
    i = regszavak(temp, &szavak);
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_TIMER;
    par->refnum = 1;
    par->ertek = par->prev = par->next = NULL;
    for (j = 0; i; i--, j++) {
	if (isalpha(szavak.s[j][0]) || szavak.s[j][0] == '$') {
	    if (nev) {
		errno = EWRONGTIMER;
		seterrtemp();
		*s = p;
		free(par);
		torol(&szavak);
#ifdef AKII_DEBUG
		dbgleave(dbg_read_timer);
#endif
		return (NULL);
	    }
	    nev = malloc(strlen(szavak.s[j]) + 1);
	    strcpy(nev, szavak.s[j]);
	}
	else if ((szavak.s[j][0] == '!') || (szavak.s[j][0] == '~')) {
	    tipus = szavak.s[j][0];
	}
	else {
	    errno = EWRONGTIMER;
	    seterrtemp();
	    *s = p;
	    free(par);
	    torol(&szavak);
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_timer);
#endif
	    return (NULL);
	}
    }
    torol(&szavak);

    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_timer);
#endif
	return (freepar(par));
    }
    i = regszavak(temp, &szavak);
    if (i = 1) {
	if (szavak.s[0][0] != '$')
	    j = strtol(szavak.s[0], &t, 10);
	else {
	    idostr = strdup(szavak.s[0] + 1);
	}
    }
    if ((i != 1) || (*t && szavak.s[0][0] != '$')) {
	errno = ETIMERWRONGTIME;
	seterrtemp();
	torol(&szavak);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_timer);
#endif
	return (freepar(par));
    }
    torol(&szavak);
    oszlop += p - mp - 1;
    inc(&p);
    ts = malloc(sizeof(struct timerstruct));

    if (!nev) {
	nev = malloc(8);
	sprintf(nev, "_%06d", (int) ((1000000.0) * rand() / (RAND_MAX + 1.0)));
    }
    ts->nev = nev;
    ts->tipus = tipus;
    if (idostr) {
	switch (tipus) {
	    case '~':
		tipus = '1';
		break;
	    default:
		tipus = '2';
	}
	ts->tipus = tipus;
	ts->ido = idostr;
    }
    else {
	ts->ido = (char *) j;
    }
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    ident++;
    if (*p == '{')
	ts->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	ts->par = readparancs(&p, PARANCS_EGY, 0);
    ident--;
    par->ertek = ts;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_timer);
#endif
    return (par);
}
struct parancs *read_on(char **s)
{
    struct parancs *par;
    struct onstruct *os;
    char *p = *s, *q = *s;
    char temp[256], *t = temp;
    REXELEM szavak =
    {0, NULL};
    int i, j;
    char *nev = NULL;
    int tipus = '~', szint = 0;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_on, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_on);
#endif
	return (NULL);
    }
    strncpy(temp, q, p - q);
    temp[p - q] = 0;
    i = regszavak(temp, &szavak);
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_ON;
    par->refnum = 1;
    par->ertek = par->prev = par->next = NULL;
    for (j = 0; i; i--, j++) {
	if (isdigit(szavak.s[j][0]) || (szavak.s[j][0] == '-')) {
	    szint = strtol(szavak.s[j], NULL, 10);
	}
	else if (isalpha(szavak.s[j][0])) {
	    if (nev) {
		errno = EWRONGON;
		seterrtemp();
		*s = p;
		free(par);
		torol(&szavak);
#ifdef AKII_DEBUG
		dbgleave(dbg_read_on);
#endif
		return (NULL);
	    }
	    nev = malloc(strlen(szavak.s[j]) + 1);
	    strcpy(nev, szavak.s[j]);
	}
	else if ((szavak.s[j][0] == '!') || (szavak.s[j][0] == '~')) {
	    tipus = szavak.s[j][0];
	}
	else {
	    errno = EWRONGON;
	    seterrtemp();
	    *s = p;
	    free(par);
	    torol(&szavak);
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_on);
#endif
	    return (NULL);
	}
    }
    torol(&szavak);

    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_on);
#endif
	return (freepar(par));
    }
    oszlop += p - mp - 1;
    inc(&p);
    i = regqszavak(temp, &szavak);
    if (i < 1) {
	errno = ENOREGKIF;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_on);
#endif
	return (freepar(par));
    }
    os = malloc(sizeof(struct onstruct));

    if (!nev) {
	nev = malloc(8);
	sprintf(nev, "_%06d", (int) ((1000000.0) * rand() / (RAND_MAX + 1.0)));
    }
    os->nev = nev;
    os->tipus = tipus;
    os->szint = szint;
    os->kifek = i;
    os->regkifek = szavak.s;
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    ident++;
    inon++;
    if (*p == '{')
	os->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	os->par = readparancs(&p, PARANCS_EGY, 0);
    inon--;
    ident--;
    par->ertek = os;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_on);
#endif
    return (par);
}
struct parancs *read_return(char **s)
{
    struct parancs *par;
    char *p = *s;
    char temp[256], *t = temp;
    struct kif *kif;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_return, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	inc(&p);
    }
    if (*p == '(') {
	mso(p);
	inc(&p);
	braces(&p, temp, '(', ')');
	if (*p != ')') {
	    errno = ENOZAROJEL;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_return);
#endif
	    return (NULL);
	}
	oszlop += p - mp - 1;
	inc(&p);
    }
    else
	t = NULL;
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_RETURN;
    par->refnum = 1;
    par->ertek = par->prev = par->next = NULL;
    if (t) {
	kif = readkif(&t);
	if (!errno)
	    par->ertek = kif;
	else {
	    free(par);
	    par = NULL;
	}
    }
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_return);
#endif
    return (par);
}
struct parancs *read_checkpoint(char **s)
{
    struct parancs *par, *par2;
    char *p = *s;
    char temp[256], *t = temp;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_checkpoint, *s);
#endif
    while (isspace(*p) && (*p))
	inc(&p);
    mso(p);
    incheck++;
    if (*p == '{')
	par2 = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	par2 = readparancs(&p, PARANCS_EGY, 0);
    incheck--;
    if (errno) {
#ifdef AKII_DEBUG
	dbgleave(dbg_read_checkpoint);
#endif
	return (NULL);
    }
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_CHECKPOINT;
    par->refnum = 1;
    par->prev = par->next = NULL;
    par->ertek = par2;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_checkpoint);
#endif
    return (par);
}
struct parancs *read_do(char **s)
{
    struct parancs *par;
    char *p = *s;
    char temp[256], *t = temp;
    struct wstruct *iew;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_do, *s);
#endif
    while (isspace(*p) && (*p))
	inc(&p);
    mso(p);
    iew = malloc(sizeof(struct wstruct));

    inloop++;
    if (*p == '{')
	iew->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	iew->par = readparancs(&p, PARANCS_EGY, 0);
    inloop--;
    if (errno) {
	free(iew);
#ifdef AKII_DEBUG
	dbgleave(dbg_read_do);
#endif
	return (NULL);
    }
    while (isspace(*p) && (*p))
	inc(&p);
    mso(p);
    while (isalpha(*p) && (*p))
	inc(&p);
    if ((mp == p) || (istrncmp(mp, "while", p - mp) != 0)) {
	errno = EWRONGDO;
	seterrtemp();
	freepar(iew->par);
	free(iew);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_do);
#endif
	return (NULL);
    }
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	if (!isspace(*p))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	freepar(iew->par);
	free(iew);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_do);
#endif
	return (NULL);
    }
    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	freepar(iew->par);
	free(iew);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_do);
#endif
	return (NULL);
    }
    oszlop += p - mp - 1;
    inc(&p);
    iew->kif = readkif(&t);
    if (errno) {
	freepar(iew->par);
	free(iew);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_do);
#endif
	return (NULL);
    }
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_DO;
    par->refnum = 1;
    par->prev = par->next = NULL;
    par->ertek = iew;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_do);
#endif
    return (par);
}
struct parancs *read_while(char **s)
{
    struct parancs *par;
    char *p = *s;
    char temp[256], *t = temp;
    struct wstruct *iew;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_while, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	if (!isspace(*p))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_while);
#endif
	return (NULL);
    }
    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_while);
#endif
	return (NULL);
    }
    oszlop += p - mp - 1;
    inc(&p);
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_WHILE;
    par->refnum = 1;
    par->prev = par->next = NULL;
    iew = malloc(sizeof(struct wstruct));

    iew->kif = readkif(&t);
    if (!errno) {
	mso(p);
	while ((*p) && (isspace(*p)))
	    inc(&p);
	inloop++;
	if (*p == '{')
	    iew->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
	else
	    iew->par = readparancs(&p, PARANCS_EGY, 0);
	inloop--;
	par->ertek = iew;
    }
    else {
	free(iew);
	free(par);
	par = NULL;
    }
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_while);
#endif
    return (par);
}
struct parancs *read_else(char **s, struct parancs **par)
{
    char *p = *s;
    char temp[256], *t = temp;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_else, *s);
#endif
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    if (*p == '{')
	(*par) = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	(*par) = readparancs(&p, PARANCS_EGY, 0);
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_else);
#endif
    return ((void *) 1);
}
struct parancs *read_if(char **s)
{
    struct parancs *par;
    char *p = *s;
    char temp[256], *t = temp;
    struct ifstruct *iew;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_if, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || (*p == ';'))
	    break;
	if (!isspace(*p))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_if);
#endif
	return (NULL);
    }
    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_if);
#endif
	return (NULL);
    }
    oszlop += p - mp - 1;
    inc(&p);
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_IF;
    par->refnum = 1;
    par->prev = par->next = NULL;
    iew = malloc(sizeof(struct ifstruct));

    iew->kif = readkif(&t);
    if (!errno) {
	mso(p);
	while ((*p) && (isspace(*p)))
	    inc(&p);
	if (*p == '{')
	    iew->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
	else
	    iew->par = readparancs(&p, PARANCS_EGY, 0);
	iew->elsepar = NULL;
	par->ertek = iew;
    }
    else {
	free(iew);
	free(par);
	par = NULL;
    }
    *s = p;
    strncpy(temp, p, 20);
    temp[20] = 0;
    while ((!isalpha(*p)) && (*p)) {
	if (*p == '#')
	    while ((*p != 10) && (*p))
		inc(&p);
	if (!isspace(*p))
	    break;
	inc(&p);
    }
    *s = p;
    if (istrncmp(p, "else", 4) == 0) {
	for (i = 0; i < 4; i++)
	    inc(&p);
	read_else(&p, &(iew->elsepar));
	par->ertek = iew;
	*s = p;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_read_if);
#endif
    return (par);
}
char **readforelements(char *p, int h)
{
    char **s;
    char *q = p;
    int a = 0, i = 0;
    s = malloc(h * (sizeof(char *)));

    while (*p) {
	if (*p == '"')
	    a ^= 1;
	if ((*p == ';') && (!a)) {
	    if (i > h - 2) {
		errno = EWRONGFOR;
		seterrtemp();
		for (i = 0; i < h; i++)
		    free(s[i]);
		free(s);
		return (NULL);
	    }
	    s[i] = malloc((p - q) + 1);
	    strncpy(s[i], q, p - q);
	    s[i][p - q] = 0;
	    i++;
	    q = ++p;
	}
	p++;
    }
    if (i == h - 1) {
	s[i] = malloc((p - q) + 1);
	strncpy(s[i], q, p - q);
	s[i][p - q] = 0;
    }
    else {
	for (; i + 1; i--)
	    free(s[i]);
	free(s);
	errno = EWRONGFOR;
	seterrtemp();
	return (NULL);
    }
    return (s);
}
struct parancs *read_for(char **s)
{
    struct parancs *par;
    char *p = *s;
    char temp[256], *t = temp;
    struct forstruct *fs;
    char **tomb;
    int i;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_for, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || !isspace(*p))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_for);
#endif
	return (NULL);
    }
    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_for);
#endif
	return (NULL);
    }
    oszlop += p - mp - 1;
    inc(&p);
    tomb = readforelements(temp, 3);
    if (!tomb) {
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_for);
#endif
	return (NULL);
    }
    fs = malloc(sizeof(struct forstruct));

    for (i = 0; i < 3; i++) {
	t = tomb[i];
	fs->kifek[i] = readkif(&t);
	if (errno) {
	    for (i = 0; i < 3; i++)
		free(tomb[i]);
	    free(tomb);
	    free(fs);
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_for);
#endif
	    return (NULL);
	}
    }
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    inloop++;
    if (*p == '{')
	fs->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	fs->par = readparancs(&p, PARANCS_EGY, 0);
    inloop--;
    par = malloc(sizeof(struct parancs));

    par->ertek = fs;
    par->tipus = PAR_FOR;
    par->refnum = 1;
    par->prev = par->next = NULL;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_for);
#endif
    return (par);
}
struct parancs *read_foreach(char **s)
{
    struct parancs *par;
    char *p = *s, *q;
    char temp[256], *t = temp;
    struct foreachstruct *fs;
    REXELEM rt =
    {0, NULL};
    char **tomb;
    int i, kettospont = 0;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_foreach, *s);
#endif
    mso(p);
    while ((*p != '(') && (*p)) {
	if ((*p == 10) || !isspace(*p))
	    break;
	inc(&p);
    }
    if (*p != '(') {
	errno = ENOZAROJELKEZD;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_foreach);
#endif
	return (NULL);
    }
    mso(p);
    inc(&p);
    braces(&p, temp, '(', ')');
    if (*p != ')') {
	errno = ENOZAROJEL;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_foreach);
#endif
	return (NULL);
    }
    oszlop += p - mp - 1;
    inc(&p);
    if (regszavak(temp, &rt) != 3) {
	torol(&rt);
	errno = EWRONGFOREACH;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_foreach);
#endif
	return (NULL);
    }
    if (istrcmp(rt.s[1], "IN") != 0 || !isalpha(rt.s[0][0])) {
	torol(&rt);
	errno = EWRONGFOREACH;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_foreach);
#endif
	return (NULL);
    }
    free(rt.s[1]);
    fs = malloc(sizeof(struct foreachstruct));

    fs->tomb = rt.s[2];
    fs->inev = rt.s[0];
    free(rt.s);
    q = fs->inev;
    while (*q)
	if (!isalnum(*q++)) {
	    q--;
	    if (*q == ':' && !kettospont) {
		kettospont = 1;
		q++;
		if (isalpha(*q))
		    continue;
	    }
	    free(fs->tomb);
	    free(fs->inev);
	    free(fs);
	    errno = EWRONGFOREACH;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_foreach);
#endif
	    return (NULL);
	}
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    inloop++;
    if (*p == '{')
	fs->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	fs->par = readparancs(&p, PARANCS_EGY, 0);
    inloop--;
    par = malloc(sizeof(struct parancs));

    par->ertek = fs;
    par->tipus = PAR_FOREACH;
    par->refnum = 1;
    par->prev = par->next = NULL;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_foreach);
#endif
    return (par);
}
struct parancs *read_alias(char **s)
{
    struct parancs *par;
    struct aliasstruct *ali;
    struct valtozo *var;
    char *q, *p = *s, *w;
    char temp[256];
    int i, j;
    REXELEM r =
    {0, NULL};

#ifdef AKII_DEBUG
    dbgenter(dbg_read_alias, *s);
#endif
    while (isspace(*p) && (*p)) {
	if (*p == 10)
	    break;
	inc(&p);
    }
    mso(p);
    if (!isalpha(*p)) {
	errno = ECSAKBETU;
	seterrtemp();
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_alias);
#endif
	return (NULL);
    }
    q = p;
    while ((!isspace(*p)) && (*p)) {
	if ((*p == 10) || (*p == '('))
	    break;
	inc(&p);
    }
    w = p;
    while (isspace(*w))
	w++;
    if (*w == '(') {
	oszlop += w - p;
	p = w;
	mso(p);
	inc(&p);
	braces(&p, temp, '(', ')');
	oszlop += p - mp - 1;
	if (*p != ')') {
	    errno = ENOZAROJEL;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_alias);
#endif
	    return (NULL);
	}
	inc(&p);
	i = regszavak(temp, &r);
	for (j = 0; j < i; j++) {
	    w = r.s[j];
	    if (!isalpha(r.s[j][0])) {
		errno = ECSAKBETU;
		seterrtemp();
		*s = p;
		torol(&r);
#ifdef AKII_DEBUG
		dbgleave(dbg_read_alias);
#endif
		return (NULL);
	    }
	    while (*w)
		if (!isalnum(*w++)) {
		    errno = ECSAKBETU;
		    seterrtemp();
		    *s = p;
		    torol(&r);
#ifdef AKII_DEBUG
		    dbgleave(dbg_read_alias);
#endif
		    return (NULL);
		}
	}
    }
    ali = malloc(sizeof(struct aliasstruct));

    ali->cpnum = 0;
    ali->cps = NULL;
    ali->nev = malloc((p - q) + 1);
    strncpy(ali->nev, q, p - q);
    ali->nev[p - q] = 0;
    q = ali->nev;
    var = getvar(&bivarlist, &q, 0, 0);
    if (var)
	if ((var->tipus == VAR_VEZER) || (var->tipus == VAR_FV)) {
	    errno = EFOGLALTVEZER;
	    seterrtemp();
	    free(ali->nev);
	    free(ali);
	    torol(&r);
#ifdef AKII_DEBUG
	    dbgleave(dbg_read_alias);
#endif
	    return (NULL);
	}
    errno = 0;
    mso(p);
    while ((*p) && (isspace(*p)))
	inc(&p);
    inalias++;
    if (*p == '{')
	ali->par = readparancs(&p, PARANCS_SOR_KAPOCS, 0);
    else
	ali->par = readparancs(&p, PARANCS_EGY, 0);
    inalias--;
    if (!ali->par) {
	if (!errno) {
	    errno = EURESALIAS;
	    seterrtemp();
	}
	free(ali->nev);
	free(ali);
	torol(&r);
	*s = p;
#ifdef AKII_DEBUG
	dbgleave(dbg_read_alias);
#endif
	return (NULL);
    }
    ali->cpnum = r.db;
    ali->cps = r.s;
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_ALIAS;
    par->refnum = 1;
    par->ertek = ali;
    par->next = par->prev = NULL;
    *s = p;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_alias);
#endif
    return (par);
}
struct parancs *read_brkcnt(char **s, int maxcount)
{
    struct parancs *par;
    char *p = *s, *q;
    int a = 0, count;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_brkcnt, *s);
#endif
    q = p;
    while (*p) {
	if (*p == '"')
	    if (*(p - 1) != '\\')
		a ^= 1;
	if ((*p == ';') && ((!a) && (*(p - 1) != '\\'))) {
	    inc(&p);
	    break;
	}
	if ((*p == '}') && ((!a) && (*(p - 1) != '\\'))) {
	    if (inkapocs)
		break;
	    else {
		errno = EZAROKAPOCS;
		mso(p);
		*s = p;
		seterrtemp();
#ifdef AKII_DEBUG
		dbgleave(dbg_read_brkcnt);
#endif
		return (NULL);
	    }
	}
	if (*p == 10)
	    break;
	inc(&p);
    }
    if (p - q) {
	count = strtol(q, NULL, 10);
	if (maxcount != -1)
	    if (count > maxcount)
		count = maxcount;
	if (count < 1)
	    count = 1;
    }
    else {
	if (maxcount == -1) {
	    count = (int) 3e7;
	}
	else
	    count = 1;
    }
    *s = p;
    par = malloc(sizeof(struct parancs));

    par->ertek = (void *) count;
    par->tipus = PAR_BREAK;
    par->refnum = 1;
    par->prev = par->next = NULL;
#ifdef AKII_DEBUG
    dbgleave(dbg_read_brkcnt);
#endif
    return (par);
}
char *stripbslash(char *s, int i, int kapocse)
{
    char *q, *p;
    int a = 0, sl = 0;

    if (!i) {
	p = malloc(1);
	*p = 0;
	return (p);
    }
    p = q = malloc(i + 1);
    for (; i; i--, s++) {
	/*if (kapocse!=PARANCS_EGY_NOVALASZ) */ if (*s == '\\') {
	    sl ^= 1;
	    if (!sl) {
		*q++ = '\\';
		/*if (((*(s+1))!=';')&&((*(s+1))!='}')) */ *q++ = '\\';
	    }
	    continue;
	}
	if (*s == '"') {
	    if (!sl)
		a ^= 1;
	}
	if (sl)
	    if ((((*s == ';') || *s == '}') && (!a))) {
		*q = *s;
		q++;
		sl = 0;
	    }
	    else {
		sl = 0;
		*q = '\\';
		q++;
		*q = *s;
		q++;
	    }
	else {
	    sl = 0;
	    *q = *s;
	    q++;
	}
    }
    if (sl)
	*q++ = '\\';
    *q = 0;
    return (p);
}
struct parancs *read_kif(char **s, int kapocse)
{
    struct parancs *par;
    struct kif *kif;
    char *p = *s, *q, *t;
    int a = 0, b = 1, sl = 0;

#ifdef AKII_DEBUG
    dbgenter(dbg_read_kif, *s);
#endif
    while (*p) {
	if (*p == '\\')
	    sl ^= 1;
	if (*p == '"')
	    if (!sl)
		a ^= 1;
	if ((*p == '{') && (!a))
	    b++;
	if ((*p == '}') && (!a))
	    b--;
	if (kapocse != PARANCS_EGY_NOVALASZ)
	    if ((*p == ';') && ((!a) && (!sl))) {
		inc(&p);
		break;
	    }
	if (inkapocs)
	    if ((*p == '}') && (!b) && ((!a) && (!sl)))
		break;
	if (*p == 10)
	    break;
	inc(&p);
	t = p;
    }
    par = malloc(sizeof(struct parancs));

    par->tipus = PAR_KIF;
    par->refnum = 1;
    par->prev = par->next = NULL;
    q = stripbslash(*s, t - *s, kapocse);
    *s = p;
    p = q;
    par->ertek = kif = readkif(&q);
    free(p);
    if (!errno) {
	if (kif->tipus == KIF_VAR) {
	    kif->tipus = KIF_FV;
	    kif->param = malloc(1);
	    kif->param[0] = 0;
	}
    }
    else {
	free(par);
	par = NULL;
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_read_kif);
#endif
    return (par);
}
struct parancs *milyenparancs(char *q, char **s, int kapocs)
{
    struct parancs *par;
    char *p = *s;
    int i = p - q, j;

#ifdef AKII_DEBUG
    dbgenter(dbg_milyenkifez, *s);
#endif
    if (istrncmp(q, "alias", i) == 0) {
	par = read_alias(&p);
	*s = p;
    }
    else if (istrncmp(q, "onsend", i) == 0) {
	par = read_on(&p);
	if (par) {
	    par->tipus = PAR_ONSEND;
	}
	*s = p;
    }
    else if (istrncmp(q, "on", i) == 0) {
	par = read_on(&p);
	*s = p;
    }
    else if (istrncmp(q, "timer", i) == 0) {
	par = read_timer(&p);
	*s = p;
    }
    else if (istrncmp(q, "if", i) == 0) {
	par = read_if(&p);
	*s = p;
    }
    else if (istrncmp(q, "do", i) == 0) {
	par = read_do(&p);
	*s = p;
    }
    else if (istrncmp(q, "while", i) == 0) {
	par = read_while(&p);
	*s = p;
    }
    else if (istrncmp(q, "foreach", i) == 0) {
	par = read_foreach(&p);
	*s = p;
    }
    else if (istrncmp(q, "for", i) == 0) {
	par = read_for(&p);
	*s = p;
    }
    else if (istrncmp(q, "break", i) == 0) {
	if (!inloop) {
	    errno = EBRKNOLOOP;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_brkcnt(&p, inloop);
	*s = p;
    }
    else if (istrncmp(q, "continue", i) == 0) {
	if (!inloop) {
	    errno = ECONTNOLOOP;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_brkcnt(&p, inloop);
	par->tipus = PAR_CONTINUE;
	*s = p;
    }
    else if (istrncmp(q, "checkpoint", i) == 0) {
	par = read_checkpoint(&p);
	*s = p;
    }
    else if (istrncmp(q, "escape", i) == 0) {
	if (!incheck) {
	    errno = EESCNOCHECK;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_brkcnt(&p, incheck);
	par->tipus = PAR_ESCAPE;
	*s = p;
    }
    else if (istrncmp(q, "nomoreon", i) == 0) {
	if (!inon) {
	    errno = ENONINON;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_brkcnt(&p, 0);
	par->tipus = PAR_NOMOREON;
	*s = p;
    }
    else if (istrncmp(q, "nexton", i) == 0) {
	if (!inon) {
	    errno = ENONINON;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_brkcnt(&p, -1);
	par->tipus = PAR_NEXTON;
	*s = p;
    }
    else if (istrncmp(q, "return", i) == 0) {
	if (!inalias) {
	    errno = ERETURNNOALIAS;
	    seterrtemp();
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_milyenkifez);
#endif
	    return (NULL);
	}
	par = read_return(&p);
	*s = p;
    }
    else {
	oszlop -= (p - q);
	par = read_kif(&q, kapocs);
	*s = q;
    }
    if (errno) {
	freepar(par);
#ifdef AKII_DEBUG
	dbgleave(dbg_milyenkifez);
#endif
	return (NULL);
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_milyenkifez);
#endif
    return (par);
}
struct parancs *readparancs(char **s, int kapocs, int doit)
{
    char *q = *s, *p = *s;
    struct parancs *par = NULL, *npar = NULL, *epar = NULL;

#ifdef AKII_DEBUG
    dbgenter(dbg_readparancs, "?");
#endif
    mso(p);
    while (*p) {
	while ((!isalpha(*p)) && (*p)) {
	    if (*p == '#')
		while ((*p != 10) && (*p))
		    inc(&p);
	    if (*p == '{') {
		if (kapocs == PARANCS_SOR_KAPOCS) {
		    inkapocs++;
		    inc(&p);
		    continue;
		}
		break;
	    }
	    if ((*p == ';') && (kapocs == PARANCS_EGY)) {
		inc(&p);
		*s = p;
#ifdef AKII_DEBUG
		dbgleave(dbg_readparancs);
#endif
		return (NULL);
	    }
	    if (*p == '}')
		if (kapocs = PARANCS_SOR_KAPOCS) {
		    inc(&p);
		    *s = p;
		    inkapocs--;
#ifdef AKII_DEBUG
		    dbgleave(dbg_readparancs);
#endif
		    return (epar);
		}
		else {
		    mso(p);
		    errno = EZAROKAPOCS;
		    seterrtemp();
		    *s = p;
#ifdef AKII_DEBUG
		    dbgleave(dbg_readparancs);
#endif
		    return (freepar(epar));
		}
	    if ((!isspace(*p)) && (*p != ';')) {
		mso(p);
		errno = EILLEGALCHAR;
		seterrtemp();
		*s = p;
#ifdef AKII_DEBUG
		dbgleave(dbg_readparancs);
#endif
		return (freepar(epar));
	    }
	    inc(&p);
	}
	mso(p);
	q = p;
	if (!*p)
	    break;
	if (*p != '{')
	    while (isalnum(*p) && (*p)) {
		inc(&p);
		if (p - q > 100) {
		    errno = EEZTULHOSSZU;
		    seterrtemp();
		    *s = p;
#ifdef AKII_DEBUG
		    dbgleave(dbg_readparancs);
#endif
		    return (freepar(epar));
		}
	    }
	else
	    inc(&p);
	if (p - q > 0) {
	    if (*q != '{')
		par = milyenparancs(q, &p, kapocs);
	    else {
		p = q;
		par = readparancs(&p, PARANCS_SOR_KAPOCS, doit);
	    }
	    if ((!par) && (errno)) {
		*s = p;
		if (msor)
		    if (strncmp(errtemp, "hiba a", 6)) {
			strcpy(errtemp + 524, errtemp);
			sprintf(errtemp, "hiba a(z) %d. sor, %d. oszlopanal: %s", msor, moszlop, errtemp + 524);
		    }
#ifdef AKII_DEBUG
		dbgleave(dbg_readparancs);
#endif
		return (freepar(epar));
	    }
	    if (!doit)
		if (!epar) {
		    epar = par;
		    while (par->next)
			par = par->next;
		    npar = par;
		}
		else {
		    npar->next = par;
		    par->prev = npar;
		    while (par->next)
			par = par->next;
		    npar = par;
		}
	    else if (par) {
		doparancssor(par);
		freepar(par);
		par = NULL;
		if ((kapocs == PARANCS_SOR_NOKAPOCS) && (purge == 1)) {
		    purgeall();
		    purge = 0;
		}
	    }
	}
	if (kapocs == PARANCS_EGY) {
	    *s = p;
#ifdef AKII_DEBUG
	    dbgleave(dbg_readparancs);
#endif
	    return (epar);
	}
    }				/*while (*p) */
    *s = p;
    if (kapocs == PARANCS_SOR_KAPOCS) {
	errno = ENOZAROKAPOCS;
	seterrtemp();
#ifdef AKII_DEBUG
	dbgleave(dbg_readparancs);
#endif
	return (freepar(epar));
    }
#ifdef AKII_DEBUG
    dbgleave(dbg_readparancs);
#endif
    return (epar);
}
